/* <GTC>
 * this file is not from the <WICED SDK DIR>
*/
/*
    FreeRTOS V7.1.0 - Copyright (C) 2011 Real Time Engineers Ltd.
	

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    >>>NOTE<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.  FreeRTOS is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/


/*
 * Implementation of pvPortMalloc() and vPortFree() that relies on the
 * compilers own malloc() and free() implementations.
 *
 * This file can only be used if the linker is configured to to generate
 * a heap memory area.
 *
 * See heap_2.c and heap_1.c for alternative implementations, and the memory
 * management pages of http://www.FreeRTOS.org for more information.
 */

#include <stdlib.h>

/* Defining MPU_WRAPPERS_INCLUDED_FROM_API_FILE prevents task.h from redefining
all the API functions to use the MPU wrappers.  That should only be done when
task.h is included from an application file. */
#define MPU_WRAPPERS_INCLUDED_FROM_API_FILE

#include "FreeRTOS.h"
#include "task.h"

#include "wiced.h"
wiced_result_t watchdog_kick( void );

#undef MPU_WRAPPERS_INCLUDED_FROM_API_FILE

/*-----------------------------------------------------------*/

#define HEAP_MIN_BLOCK_SIZE 16

typedef enum 
{
	BEST_FIT = 1,
	WORST_FIT = 2,
	FIRST_FIT = 3,
}
heap_search_algorithm_t;

#define NUM_HEAP_INDEX 4

typedef struct 
{
	void *heap_start;
	int heap_start_size;
	int heap_free;
	unsigned int heap_count;
	short heap_index_count[NUM_HEAP_INDEX];
	unsigned int heap_flag_init : 1;
	heap_search_algorithm_t search_algorithm : 2;
	unsigned int overflow_check_on : 1;
}
heap_controls_t;

typedef struct
{
	unsigned int back_size : 31;
	unsigned int used : 1;
	unsigned int forward_size : 32;
	#if (HEAP_USE_MAGIC_NUMBER)
	unsigned int magic_number;
	#endif
}
heap_header_t;


static heap_controls_t heap_controls;

extern int _heap_;

void *pvPortMalloc_withIndex( size_t xWantedSize, unsigned char index )
{
	heap_controls.heap_index_count[index]++;
	return pvPortMalloc(xWantedSize);
}

void *pvPortMalloc( size_t xWantedSize )
{
	void *pvReturn = NULL;
	
	watchdog_kick( );
	
	// align the memory to 4-bytes
	if( xWantedSize & 0x3 )
	{
		xWantedSize >>= 2;
		xWantedSize++;
		xWantedSize <<= 2;
	}
	int want_total = xWantedSize + sizeof(heap_header_t);
	
	// initialize
	if(!heap_controls.heap_flag_init)
	{
		
		heap_controls.heap_start_size = (int)&_heap_;
		
		vTaskSuspendAll();
		do{
			heap_controls.heap_start = malloc( --heap_controls.heap_start_size );
		} while (!heap_controls.heap_start);
		xTaskResumeAll();
		
		heap_controls.heap_free = heap_controls.heap_start_size;
		heap_controls.heap_count = 0;
		
		if(!heap_controls.heap_start)
			asm("nop");
		
		heap_controls.search_algorithm = FIRST_FIT;
		heap_controls.overflow_check_on = 0;
		if(heap_controls.overflow_check_on)
			memset(heap_controls.heap_start, 0x00, heap_controls.heap_start_size);
		
		((heap_header_t*)(heap_controls.heap_start))->back_size = 0;
		((heap_header_t*)(heap_controls.heap_start))->used = 0;
		((heap_header_t*)(heap_controls.heap_start))->forward_size = heap_controls.heap_start_size;
		
		#if (HEAP_USE_MAGIC_NUMBER)
		((heap_header_t*)(heap_controls.heap_start))->magic_number = 0x78563412;
		#endif
		
		heap_controls.heap_flag_init = 1;
		
	}

	// search algorithm
	vTaskSuspendAll();
	switch(heap_controls.search_algorithm)
	{
	case BEST_FIT:
		{
			// NYI
			BREAK_AND_LOG(LOG_HEAP_X);
			break;
		}
	case WORST_FIT:
		{
			// NYI
			BREAK_AND_LOG(LOG_HEAP_X);
			break;
		}
	case FIRST_FIT:
		{
			heap_header_t * ptr;
			retry:
			// Start at top of heap, scan until we find an unused block that is big enough
			ptr = heap_controls.heap_start;
			int total_size = 0;
			int total_count = 0;
			
			while( (int)ptr < ((int)(heap_controls.heap_start) + heap_controls.heap_start_size) )
			{
				watchdog_kick( );
				// is the slot vacant?
				if( ptr->used )
				{
					int delta = ((heap_header_t*)ptr)->forward_size;
					if(delta)
					{
						ptr = (void*)((int)ptr + delta);
						
						total_size += delta;
						total_count++;
						if(ptr->back_size != delta)
						{
							// internal heap structure is broken
							BREAK_AND_LOG(LOG_HEAP_X);
						}
						continue;
					}
					else
					{
						BREAK_AND_LOG(LOG_HEAP_X);
						break;
					}
				}
				else
				{
					if(heap_controls.overflow_check_on)
					{
						// verify no overflows
						for(int i = 0; i < ((heap_header_t*)ptr)->forward_size - sizeof(heap_header_t); i++)
						{
							char * p = (char*)ptr + sizeof(heap_header_t);
							if(*p != 0x00)
							{
								// someone corrupted this memory
								BREAK_AND_LOG(LOG_HEAP_X);
							}
						}
					}
				}
				
				// is the slot big enough?
				if( ((heap_header_t*)ptr)->forward_size >= (want_total) )
				{
					pvReturn = ptr;
					break;
				}
				else
				{
					if(((heap_header_t*)ptr)->forward_size)
					{
						ptr = (void*)((int)ptr + ((heap_header_t*)ptr)->forward_size);
						continue;
					}
					else
					{
						BREAK_AND_LOG(LOG_HEAP_X);
						break;
					}
				}
			}
			
			if(pvReturn)
				break;
			else
			{
				int retry = 0;
				
				BREAK_AND_LOG(LOG_HEAP_X);
				
				if(retry)
					goto retry;
				else
					break;
			}
		}
	default:
		break;
	}
	
	// assign
	
	if( pvReturn )
	{
		// if fit is 'tight', assign the whole block, else break the block up
		if( (((heap_header_t*)pvReturn)->forward_size - want_total) < HEAP_MIN_BLOCK_SIZE )
		{
			((heap_header_t*)pvReturn)->used = 1;
			// this block's forward_size is already set
			// next block's back_size should already be set
		}
		else
		{
			heap_header_t *new_next_block, *old_next_block;
			
			old_next_block = (void*)((int)pvReturn + ((heap_header_t*)pvReturn)->forward_size);
			new_next_block = (void*)((int)pvReturn + want_total);
			
			if( (int)old_next_block >= (int)(heap_controls.heap_start) + heap_controls.heap_start_size )
				old_next_block = NULL;
			
			((heap_header_t*)pvReturn)->used = 1;
			
			// update old next block's back_size
			if(old_next_block)
				old_next_block->back_size -= want_total;
			
			// update new next block's back_size
			new_next_block->back_size = want_total;
			
			// update new next block's forward_size
			new_next_block->forward_size = ((heap_header_t*)pvReturn)->forward_size - want_total;
			
			// update new next block's used field (clear it)
			new_next_block->used = 0;
			
			#if (HEAP_USE_MAGIC_NUMBER)
			new_next_block->magic_number = 0x78563412;
			#endif
			
			// update current block's forward_size
			((heap_header_t*)pvReturn)->forward_size = want_total;
			
			
			// validate sizes
			if( old_next_block && new_next_block && old_next_block->back_size != new_next_block->forward_size )
			{
				BREAK_AND_LOG(LOG_HEAP_X);
			}
			if( new_next_block && new_next_block->back_size != ((heap_header_t*)pvReturn)->forward_size )
			{
				BREAK_AND_LOG(LOG_HEAP_X);
			}
		}
		
		heap_controls.heap_free -= ((heap_header_t*)pvReturn)->forward_size;
		
		// shift pointer by header
		pvReturn = (void*)((int)pvReturn + sizeof(heap_header_t));
		
		// validate, nothing went crazy
		if ( (int)pvReturn < (int)heap_controls.heap_start
			|| (int)pvReturn > ( (int)heap_controls.heap_start + heap_controls.heap_start_size) )
		{
			BREAK_AND_LOG(LOG_HEAP_X);
		}
	}
	xTaskResumeAll();
	
	// return
	#if( configUSE_MALLOC_FAILED_HOOK == 1 )
	if( pvReturn == NULL )
	{
		extern void vApplicationMallocFailedHook( void );
		vApplicationMallocFailedHook();
	}
	#endif
	
	if( pvReturn )
		heap_controls.heap_count++;
	
	return pvReturn;
}
/*-----------------------------------------------------------*/

void vPortFree_withIndex( void *pv, unsigned char index )
{
	heap_controls.heap_index_count[index]--;
	vPortFree(pv);
	return;
}

void vPortFree( void *pv )
{
	if( pv )
	{
		heap_header_t *cur_block = NULL, *back_block = NULL, *forward_block = NULL;
		char *cleanup = NULL;
		int size;
		
		cur_block = (void*)((int)pv - sizeof(heap_header_t));
		
		size = cur_block->forward_size;
		cleanup = (void*)cur_block;
		
		// bounds check our input
		if( (int)cur_block < (int)heap_controls.heap_start
			|| (int)cur_block > ((int)heap_controls.heap_start + heap_controls.heap_start_size) )
		{
			// input is outside valid range
			BREAK_AND_LOG(LOG_HEAP_X);
			return;
		}
		
		// verify magic number did not get corrupted
		#if (HEAP_USE_MAGIC_NUMBER)
		if( cur_block->magic_number != 0x78563412 )
		{
			BREAK_AND_LOG(LOG_HEAP_X);
		}
		#endif
		
		// find the back_block
		if(cur_block->back_size)
		{
			back_block = (void*)((int)cur_block - cur_block->back_size);
			
			if( cur_block->back_size != back_block->forward_size )
			{
				BREAK_AND_LOG(LOG_HEAP_X);
			}
		}
		
		// find the forward_block
		if(cur_block->forward_size)
		{
			forward_block = (void*)((int)cur_block + cur_block->forward_size);
			
			if( cur_block->forward_size != forward_block->back_size )
			{
				BREAK_AND_LOG(LOG_HEAP_X);
			}
		}
		
		// if back block is free, merge this block with it
		if(back_block && !back_block->used)
		{
			back_block->forward_size += cur_block->forward_size;
			if(forward_block)
				forward_block->back_size += cur_block->back_size;
			
			cleanup = (void*)back_block;
		}
		// otherwise, leave this block, but mark it as unused
		else
			cur_block->used = 0;
		
		// also check to see if we can merge with the forward_block
		if(forward_block && !forward_block->used)
		{
			heap_header_t *forward_forward_block = NULL;
			if(forward_block->forward_size)
				forward_forward_block = (void*)((int)forward_block + forward_block->forward_size );
			if( (int)forward_forward_block > ((int)heap_controls.heap_start + heap_controls.heap_start_size) )
				forward_forward_block = NULL;
			if(forward_forward_block)
				forward_forward_block->back_size += forward_block->back_size;
			
			// triple merge?
			if(back_block && !back_block->used)
			{
				back_block->forward_size += forward_block->forward_size;
				
				cleanup = (void*)back_block;
			}
			else
			{
				// simple forward merge
				cur_block->forward_size += forward_block->forward_size;
				
				cleanup = (void*)cur_block;
			}
		}
		
		if(heap_controls.overflow_check_on)
		{
			if(cleanup)
			{
				memset( cleanup + sizeof(heap_header_t), 0x00, ((heap_header_t*)cleanup)->forward_size - sizeof(heap_header_t) );
			}
			else
				// no cleanup!
				BREAK_AND_LOG(LOG_HEAP_X);
		}
		
		heap_controls.heap_free += size;
		heap_controls.heap_count--;
	}
}



